/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.Record;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.SearchMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaData;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opentest4j.TestAbortedException;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mappers.mapper;
import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaDataBuilder.tableMetaData;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.Mockito.when;

@ExtendWith(MockitoExtension.class)
class ReadersTest {

    @NonNull private static final String CATALOG_NAME = "catalog";
    @NonNull private static final String SCHEMA_NAME = "schema";
    @NonNull private static final String TABLE_NAME = "table";
    @NonNull private static final TableMetaData TABLE_META_DATA = tableMetaData(TABLE_NAME)
        .schemaName(SCHEMA_NAME)
        .catalogName(CATALOG_NAME)
        .build();
    @NonNull private static final String KEY_COLUMN_LABEL = "key";
    @NonNull private final DataSource dataSource;
    @NonNull private final AtomicInteger numberOfResultSets;
    @NonNull private final ResultSet resultSet;

    ReadersTest(@Mock final DataSource dataSource,
                @Mock final Connection connection,
                @Mock final ResultSet resultSet,
                @Mock final PreparedStatement preparedStatement) {
        this.dataSource = dataSource;
        this.resultSet = resultSet;
        numberOfResultSets = new AtomicInteger();
        try {
            when(resultSet.next()).thenAnswer(i -> numberOfResultSets.getAndDecrement() > 0);
            when(dataSource.getConnection()).thenReturn(connection);
            when(connection.prepareStatement("select * from " + TABLE_NAME + " where " + KEY_COLUMN_LABEL + " = ?")).thenReturn(preparedStatement);
            when(preparedStatement.executeQuery()).thenReturn(resultSet);
        } catch (Exception e) {
            throw new TestAbortedException("Failed to hook mocks",e);
        }

    }

    @Test
    void keyReader() throws SQLException {
        final KeyMapper<MyRecord, String> mapper = mapper(
            String.class,
            TABLE_META_DATA,
            TABLE_META_DATA,
            MyRecord::new)
            .addKeyStringMapping(KEY_COLUMN_LABEL, MyRecord::key, MyRecord::key, Function.identity())
            .build();

        final Reader<MyRecord, String> reader = Readers.reader(dataSource, mapper);
        final String hit = hit("hit");
        final MyRecord myRecord = reader.search("");
        assertNotNull(myRecord);
        assertEquals(hit, myRecord.key);
        miss("miss");
        assertNull(reader.search(""));
    }

    @Test
    @SuppressWarnings("ConstantConditions")
    void searchReader() throws SQLException {
        final SearchMapper<MyRecord, String> mapper = mapper(
            String.class,
            TABLE_META_DATA,
            TABLE_META_DATA,
            MyRecord::new)
            .addSearchStringMapping(KEY_COLUMN_LABEL, MyRecord::key, MyRecord::key, Function.identity())
            .build();

        final Reader<Collection<MyRecord>, String> reader = Readers.reader(dataSource, mapper);
        final String[] hits = hits("a", "b", "c");
        final Collection<MyRecord> records = reader.search("");
        final Iterator<MyRecord> iterator = records.iterator();
        for(int i = 0; i < hits.length && iterator.hasNext(); i++) {
            MyRecord record = iterator.next();
            assertNotNull(record);
            assertEquals(hits[i], record.key);
        }
        final String miss = miss("");
        assertTrue(reader.search(miss).isEmpty());
    }

    @Test
    void recordReader(
        @Mock final KeyMapper<Record<String, String>, String> mapper,
        @Mock final Record<String, String> record) throws SQLException {
        String hit = hit("x");
        when(mapper.selectTableMetaData()).thenReturn(TABLE_META_DATA);
        when(mapper.keyColumnLabel()).thenReturn(KEY_COLUMN_LABEL);
        when(mapper.serializeKey(any())).then(i -> i.getArgument(0, String.class));
        when(mapper.mapRow(any(ResultSet.class), anyInt())).thenReturn(record);
        when(record.data()).thenReturn(hit);
        final Reader<String, String> reader = Readers.recordReader(dataSource, mapper);
        final String data = reader.search("");
        assertEquals(hit, data);
        miss("x");
        assertNull(reader.search(""));
    }

    @NonNull
    private String hit(@NonNull final String key) throws SQLException {
        return hits(key)[0];
    }

    @NonNull
    private String[] hits(@NonNull final String... keys) throws SQLException {
        numberOfResultSets.set(keys.length);
        when(resultSet.getString(KEY_COLUMN_LABEL)).then(i -> keys[keys.length - numberOfResultSets.get() - 1]);
        return keys;
    }

    @NonNull
    private String miss(@NonNull final String key) {
        numberOfResultSets.set(0);
        return key;
    }

    private static final class MyRecord {
        private String key;

        private String key() {
            return key;
        }

        @NonNull
        private void key(@NonNull final String key) {
            this.key = key;
        }
    }

}
