/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import com.google.common.base.Strings;
import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.LabResultRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Extractor;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.LabResultETL;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Transformer;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DataFetcherException;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.KidsDataException;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultImplBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.kids.dao.KidsLabResult;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Lazy;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.Collections;
import java.util.function.BiFunction;
import java.util.function.Supplier;

@Configuration
public class KIDSConfig {
    private static final Logger logger = Logger.getLogger(KIDSConfig.class);
    public static final String POSITIV = "POSITIV";
    public static final String NEGATIV = "NEGATIV";
    private final int labResultsRecordsTimeToLiveDays;

    public KIDSConfig(@Value("${datafetcher.lab-result.records.time-to-live-days:-1}") int labResultsRecordsTimeToLiveDays) {
        this.labResultsRecordsTimeToLiveDays = labResultsRecordsTimeToLiveDays;
    }

    @Bean
    @Lazy
    @LabResultETL
    @NonNull
    public Extractor<Collection<KidsLabResult>> extractor(@NonNull final Reader<Collection<KidsLabResult>, String> kidsReader) {
        return cpr -> {
            if (StringUtils.isBlank(cpr))
                throw new DataFetcherException("CPR number is missing");
            Collection<KidsLabResult> kidsLabResults = kidsReader.search(cpr);
            return kidsLabResults != null ? kidsLabResults : Collections.emptyList();
        };
    }

    @Bean
    @Lazy
    @LabResultETL
    @NonNull
    public Transformer<Collection<KidsLabResult>, LabResultRecord<LabResults, Collection<KidsLabResult>>> kidsTransformer(
        @NonNull final BiFunction<LabResults, Collection<KidsLabResult>, LabResultRecord<LabResults, Collection<KidsLabResult>>> labResultsRecordTransformer) {
        return (cpr, source) -> {
            if (StringUtils.isBlank(cpr))
                throw new DataFetcherException("CPR number is missing");
            LabResults dto = new LabResultsImpl(cpr);
            for (KidsLabResult kidsLabResult : source) {
                if (!cpr.equals(kidsLabResult.cpr()))
                    throw new DataFetcherException("Input cpr does not match KIDS record cpr");
                String result = kidsLabResult.result();
                if (!StringUtils.equalsAnyIgnoreCase(result, POSITIV, NEGATIV)) {
                    throw new KidsDataException("KidsLabResult inconclusive (was not '" + POSITIV + "' or '" + NEGATIV + "'). Result=" + result);
                }
                if (StringUtils.isBlank(kidsLabResult.type())) {
                    throw new KidsDataException("KidsLabResult test type missing");
                }
                final LabResultImplBuilder builder = LabResultImplBuilder.labResult(
                    kidsLabResult.id(),
                    result.equalsIgnoreCase(POSITIV),
                    kidsLabResult.type(),
                    kidsLabResult.sampleTime(),
                    kidsLabResult.resultTime());

                builder
                    .name(Strings.emptyToNull(kidsLabResult.name()))
                    .manufacturer(Strings.emptyToNull(kidsLabResult.manufacturer()))
                    .center(Strings.emptyToNull(kidsLabResult.center()));
                boolean added = dto.add(builder.build());
                if (!added) {
                    throw new KidsDataException("Identical KIDS records detected for this cpr. Cannot proceed");
                }
            }
            return labResultsRecordTransformer.apply(dto, source);
        };
    }

    @Bean
    @Lazy
    @NonNull
    public BiFunction<LabResults, Collection<KidsLabResult>, LabResultRecord<LabResults, Collection<KidsLabResult>>> labResultsRecordProducer(
        @NonNull Supplier<LabResultRecordImpl<LabResults, Collection<KidsLabResult>>> labResultRecordConstructor) {
        return (data, source) -> labResultRecordConstructor.get()
            .cpr(data.cpr())
            .creationTime(LocalDateTime.now())
            .traceId("TODO")
            .setData(data)
            .invalidationTime(labResultsRecordsTimeToLiveDays > 0 ? LocalDateTime.now().plusDays(labResultsRecordsTimeToLiveDays) : null)
            .setSource(source);
    }
}
