/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.mocks;

import dk.oio.rep.cpr_dk.xml.schemas.core._2006._01._17.SimpleCPRPersonType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Extractor;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.VaccinationETL;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import dk.vaccinationsregister.schemas._2013._12._01.VaccinationStatus;
import org.apache.log4j.Logger;
import org.springframework.context.annotation.*;
import org.springframework.lang.NonNull;

import java.util.Random;

@Configuration
public class MockConfiguration {
    private static final Logger log = Logger.getLogger(MockConfiguration.class);

    /**
     * Mock Extractor for DDV
     */
    @Bean
    @Lazy
    @Primary
    @VaccinationETL
    @Profile("mock-ddv")
    @NonNull
    public Extractor<GetVaccinationPassportAsPDFResponseType> mockDdvExtractor() {
        log.info("Using Mock DDV");
        Random random = new Random();
        double meanSleep = 0;
        double variance = 0;
        if (System.getenv().containsKey("MOCK_DDV_SLEEP_MILLIS")) {
            String sleepTime = System.getenv("MOCK_DDV_SLEEP_MILLIS");
            try {
                meanSleep = Long.parseLong(sleepTime);
                variance = meanSleep / 2;
            } catch (NumberFormatException numberFormatException) {
                log.error("Not doing Mock DDV sleep: Could not parse environment variable MOCK_DDV_SLEEP_MILLIS to a Long: " + sleepTime);
            }
        }

        double finalMeanSleep = meanSleep;
        double finalVariance = variance;
        return cpr -> {
            log.info("Calling Mock DDV");
            if (finalMeanSleep > 0) {
                double randomSleep = finalMeanSleep + random.nextGaussian() * finalVariance;
                try {
                    Thread.sleep((long) Math.max(0, randomSleep));
                } catch (InterruptedException interruptedException) {
                    log.error("Mock DDV sleep interrupted");
                    Thread.currentThread().interrupt();
                }
            }

            GetVaccinationPassportAsPDFResponseType response = new GetVaccinationPassportAsPDFResponseType();
            response.setVaccinationAgainst("SARS-CoV-2 (COVID-19)");
            response.setVaccinationStatus(VaccinationStatus.VACCINERET);
            SimpleCPRPersonType pi = new SimpleCPRPersonType();
            pi.setPersonCivilRegistrationIdentifier(cpr);
            PersonNameStructureType pn = new PersonNameStructureType();
            pn.setPersonGivenName("Lars");
            pn.setPersonMiddleName("Linde");
            pn.setPersonSurnameName("Larsen");
            pi.setPersonNameStructure(pn);
            response.setPersonInfo(pi);
            response.setDocumentId("100000007");
            return response;
        };
    }
}
