/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonGetter;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.io.IOException;
import java.time.LocalDateTime;
import java.util.Comparator;
import java.util.List;

import static java.util.Collections.unmodifiableList;
import static java.util.Comparator.naturalOrder;

public class VaccinationImpl implements Vaccination {

    private static final String CPR_NAME_PROPERTY = "cpr";
    private static final String VACCINATIONS_PROPERTY = "vaccinations";
    private static final String COMPLETED_PROPERTY = "completed";

    @NonNull private final String cpr;
    @NonNull private final List<EffectuatedVaccination> vaccinations;
    @Nullable private final LocalDateTime completed;
    @Nullable private final EffectuatedVaccination latestEffectuated;

    @JsonCreator
    VaccinationImpl(
        @NonNull @JsonProperty(CPR_NAME_PROPERTY) final String cpr,
        @NonNull @JsonProperty(COMPLETED_PROPERTY) boolean completed,
        @NonNull @JsonProperty(VACCINATIONS_PROPERTY) final List<EffectuatedVaccination> vaccinations) {
        this.cpr = cpr;
        this.vaccinations = unmodifiableList(vaccinations);
        this.completed = completed ? vaccinations.stream()
            .map(EffectuatedVaccination::effectuated)
            .max(naturalOrder())
            .orElse(null) : null;
        this.latestEffectuated = vaccinations.stream().max(Comparator.comparing(EffectuatedVaccination::effectuated)).orElse(null);

    }


    @Nullable
    public EffectuatedVaccination getLatestEffectuated() {
        return latestEffectuated;
    }

    @JsonGetter
    @JsonProperty(CPR_NAME_PROPERTY)
    @Override
    @NonNull
    public String cpr() {
        return cpr;
    }

    @Override
    @Nullable
    @JsonGetter
    @JsonProperty(COMPLETED_PROPERTY)
    @JsonSerialize(nullsUsing = CompletedSerializer.class, using = CompletedSerializer.class)
    public LocalDateTime completed() {
        return completed;
    }

    @JsonGetter
    @JsonProperty(VACCINATIONS_PROPERTY)
    @Override
    @NonNull
    public List<EffectuatedVaccination> vaccinations() {
        return vaccinations;
    }

    public static class CompletedSerializer extends StdSerializer<LocalDateTime> {

        protected CompletedSerializer() {
            super(LocalDateTime.class);
        }

        @Override
        public void serialize(LocalDateTime value, JsonGenerator gen, SerializerProvider provider) throws IOException {
            gen.writeBoolean(value != null);
        }
    }

}
