/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.module.SimpleModule;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;
import static java.util.Comparator.naturalOrder;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

class VaccinationImplTest {

    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());
    private final ObjectMapper objectMapper;

    VaccinationImplTest() {
        final SimpleModule module = new SimpleModule()
            .addAbstractTypeMapping(EffectuatedVaccination.class, EffectuatedVaccinationImpl.class)
            .addAbstractTypeMapping(Vaccination.class, VaccinationImpl.class);
        objectMapper = new ObjectMapper().findAndRegisterModules().registerModule(module).disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);
    }

    @ParameterizedTest
    @MethodSource("sources")
    @DisplayName("Invariant: If there is a completed date and time, it is the date and time of the most recently effectuated vaccination.")
    void completedAndVaccinations(@NonNull final Vaccination vaccination) {
        if(completed(vaccination)) assertTrue(thereAreVaccinations(vaccination));
        if(completed(vaccination)) assertEquals(vaccination.completed(), mostRecentVaccinationDate(vaccination));
        if(!vaccination.vaccinations().isEmpty())
        assertEquals(mostRecentVaccinationDate(vaccination), vaccination.getLatestEffectuated().effectuated());
    }

    @ParameterizedTest
    @MethodSource("sources")
    @DisplayName("Serialize and deserialze")
    void serializeDeserialize(@NonNull final Vaccination source) throws JsonProcessingException {

        final String serialized = objectMapper.writeValueAsString(source);
        assertNotNull(serialized);

        final Vaccination deserialized = objectMapper.readValue(serialized, Vaccination.class);
        assertNotNull(deserialized);
        assertEquals(source.cpr(), deserialized.cpr());
        assertEquals(source.completed(), deserialized.completed());
        assertEquals(source.vaccinations().size(), deserialized.vaccinations().size());
    }

    private static boolean completed(@NonNull Vaccination vaccination) {
        return vaccination.completed() != null;
    }

    private LocalDateTime mostRecentVaccinationDate(Vaccination vaccination) {
        return vaccination.vaccinations().stream()
            .map(EffectuatedVaccination::effectuated)
            .max(naturalOrder())
            .orElse(null);
    }

    private static boolean thereAreVaccinations(Vaccination vaccination) {
        return !vaccination.vaccinations().isEmpty();
    }

    @NonNull
    private static Stream<Arguments> sources() {
        return generate(VaccinationImplTest::newVaccination)
            .limit(100)
            .map(Arguments::arguments);
    }

    @NonNull
    private static Vaccination newVaccination() {
        final VaccinationBuilder builder = vaccination(RANDOM_DATA.cpr(), RANDOM_DATA.nextBoolean());
        generate(VaccinationImplTest::newEffectuatedVaccination)
            .limit(RANDOM_DATA.nextInt(10))
            .forEach(builder::add);
        return builder.build();
    }

    @NonNull
    private static EffectuatedVaccination newEffectuatedVaccination() {
        return effectuatedVaccination(RANDOM_DATA.nextLong(), RANDOM_DATA.string(10), RANDOM_DATA.string(10))
            .effectuated(RANDOM_DATA.localDateTime())
            .build();
    }

}
