/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.security;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.cbs.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.SecurityRuntimeException;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import java.util.Optional;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.when;

@CbsTest
@SuppressWarnings("java:S5979") /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
class SecurityTest {
    @Mock
    private SecurityContext.Ticket ticket;
    @Mock
    private SecurityContext.User user;
    @Mock
    private SecurityContext securityContext;
    @Mock
    private GetCoronaPassRequest request;
    private final Security security = new Security();

    @Test
    void throws_exception_when_no_ticket() {
        when(securityContext.getTicket()).thenReturn(Optional.empty());
        SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> security.checkSecurityTicketGetCpr(securityContext, request));
        assertEquals(ErrorMessages.NO_SECURITY_TICKET, exception.getMessage());
    }

    @Test
    void throws_exception_on_invalid_ticket() {
        when(ticket.isValid()).thenReturn(false);
        when(securityContext.getTicket()).thenReturn(Optional.of(ticket));
        SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> security.checkSecurityTicketGetCpr(securityContext, request));
        assertEquals(ErrorMessages.INVALID_SECURITY_TICKET, exception.getMessage());
    }

    @Test
    void throws_exception_on_cpr_mismatch() {
        String cpr2 = "0202023333";
        String cpr1 = "0101012222";

        when(ticket.isValid()).thenReturn(true);
        when(securityContext.getTicket()).thenReturn(Optional.of(ticket));
        when(securityContext.getActingUser()).thenReturn(Optional.of(user));
        when(user.getIdentifier()).thenReturn(cpr1);
        when(request.getIdentifier()).thenReturn(cpr2);

        SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> security.checkSecurityTicketGetCpr(securityContext, request));
        assertEquals(ErrorMessages.NOT_AUTHORIZED_CPR_MISMATCH, exception.getMessage());
    }
}
