/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultImpl;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import se.digg.dgc.payload.v1.Tst;

import java.time.ZoneOffset;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultImplBuilder.labResult;
import static org.junit.jupiter.api.Assertions.*;

public class LabResultMapperTest {

    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());
    private LabResultMapper labResultMapper;


    @BeforeEach
    void beforeEach() {
        labResultMapper = new LabResultMapper();
    }

    @Test
    void testMapOfResults_unknownLabResultType() {
        LabResultImpl labResult = labResult(
            RANDOM_DATA.nextLong(),
            RANDOM_DATA.nextBoolean(),
            "q",
            RANDOM_DATA.localDateTime(),
            RANDOM_DATA.localDateTime())
            .center(RANDOM_DATA.nextBoolean() ? RANDOM_DATA.string(10) : null)
            .manufacturer(RANDOM_DATA.nextBoolean() ? RANDOM_DATA.string(10) : null)
            .build();
        Exception exception = assertThrows(LabResultMapperException.class, () -> labResultMapper.toTst(labResult));
        assertTrue(exception.getMessage().contains("Unknown LabResult.type: 'q' - labresult cannot be mapped"));
    }

    @Test
    void testMapOfLabResultPCR() {
        LabResultImpl labResult = labResult(
            RANDOM_DATA.nextLong(),
            true,
            "SARS2",
            RANDOM_DATA.localDateTime(),
            RANDOM_DATA.localDateTime())
            .center(RANDOM_DATA.nextBoolean() ? RANDOM_DATA.string(10) : null)
            .manufacturer(RANDOM_DATA.nextBoolean() ? RANDOM_DATA.string(10) : null)
            .build();

        Tst tst = labResultMapper.toTst(labResult);
        assertEquals("DK", tst.getCou());
        assertEquals(LabResultMapper.LOINC_PCR, tst.getTyp());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, tst.getDis());
        assertEquals(labResult.center(), tst.getFac());
        assertEquals(labResult.manufacturer(), tst.getTma());
        assertEquals(LabResultMapper.SNOMED_CT_DETECTED_POSITIVE, tst.getRes());
        assertEquals(labResult.resultDateTime().toEpochSecond(ZoneOffset.UTC), tst.getDtr().intValue());
        assertEquals(labResult.samplingDateTime().toEpochSecond(ZoneOffset.UTC), tst.getDts().intValue());
        assertEquals(labResult.name(), tst.getOri());
    }

    @Test
    void testMapOfLabResultAntigen() {
        LabResultImpl labResult = labResult(
            RANDOM_DATA.nextLong(),
            true,
            "SARSG",
            RANDOM_DATA.localDateTime(),
            RANDOM_DATA.localDateTime())
            .center(RANDOM_DATA.nextBoolean() ? RANDOM_DATA.string(10) : null)
            .manufacturer(RANDOM_DATA.nextBoolean() ? RANDOM_DATA.string(10) : null)
            .build();

        Tst tst = LabResultMapper.toTst(labResult);
        assertEquals("DK", tst.getCou());
        assertEquals(LabResultMapper.LOINC_ANTIGEN, tst.getTyp());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, tst.getDis());
        assertEquals(labResult.center(), tst.getFac());
        assertEquals(labResult.manufacturer(), tst.getTma());
        assertEquals(LabResultMapper.SNOMED_CT_DETECTED_POSITIVE, tst.getRes());
        assertEquals(labResult.resultDateTime().toEpochSecond(ZoneOffset.UTC), tst.getDtr().intValue());
        assertEquals(labResult.samplingDateTime().toEpochSecond(ZoneOffset.UTC), tst.getDts().intValue());
        assertEquals(labResult.name(), tst.getOri());
    }

}
