/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.LibertyResponse;
import dk.sosi.seal.vault.CredentialVault;
import dk.sundhedsdatastyrelsen.cbs.configuration.BorgerContext;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.configuration.RequestContext;
import dk.sundhedsdatastyrelsen.cbs.configuration.XmlDocumentResponseInterceptor;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.CoronaPassPort;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.w3c.dom.Document;

import javax.xml.bind.JAXB;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.MessageContext;
import java.io.ByteArrayOutputStream;
import java.util.Collections;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

@ExtendWith(SpringExtension.class)
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
class CoronaPassITCase {
    private static final Logger log = Logger.getLogger(CoronaPassITCase.class);
    private static final String FMK_AUDIENCE = "https://fmk";
    private static final String IDWS_CPR = "3108732076";
    private static final String IDWS_PID = "PID:9208-2002-2-098076399641";
    private String REQUEST_CPR = "3108732076";

    @Autowired
    CoronaPassPort servicePort;
    @Autowired
    BorgerContext borgerContext;
    @Autowired
    RequestContext requestContext;
    @Autowired
    XmlDocumentResponseInterceptor documentInterceptor;
    @Autowired
    CredentialVault trustVault;
    @Autowired
    OIOIDWSFactory oioIdwsFactory;

    @BeforeEach
    public void setup() {
        borgerContext.clearContext();
        requestContext.clearContext();
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_same_cpr_x_times() {
        for (int i = 0; i < 10; i++) {
            testGetCoronaPasses();
        }
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_different_cpr_x_times() {
        RandomData randomData = new RandomData(System.currentTimeMillis());
        for (int i = 0; i < 2; i++) {
            REQUEST_CPR = randomData.cpr();
            testGetCoronaPasses();
        }
    }

    @Test
    void testGetCoronaPasses() {
        /*
          DDV test1
          2401010010: Pas OK(Pfizer)
          2401010011: Pas ej OK, der er kun givet 1 vaccination
          2401010012: Pas OK (Moderna)
          2401010013: Pas OK (Pfizer)
          2401010014: Pas ej OK (min.interval ikke overholdt)
          2401010015: Pas ej OK (kun 1 vaccination)
          2401010016: Pas OK (Pfizer)
          2401010017: Pas OK (Pfizer)
          2401010018: Pas OK (krydsforløb med AstraZeneca + Moderna)

          insert-kids-data.sql (note that '2401010010' has no data):
          (person_identifier, test_identifier, type_of_test, test_name, test_manufacturer, timestamp_sample, timestamp_result, result_of_the_test, testing_centre_or_facility, timestamp_last_update, sletteflag)
          ('2401010011', 2021001016099, 'SARS2', '', '', '2021-01-12 10:29:00', '2021-01-12 20:19:00', 'NEGATIV', '', '2021-04-22 03:51:37', 0),
          ('2401010012', 2021010110318, 'SARS2', '', '', '2021-03-11 09:15:00', '2021-04-23 10:40:00', 'NEGATIV', '', '2021-04-23 11:23:40', 0),
          ('2401010012', 2021013666631, 'SARSG', '', '', '2021-02-20 10:30:00', '2021-02-20 10:50:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010013', 2021013966213, 'SARSG', '', '', '2021-03-23 07:09:00', '2021-03-23 07:29:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010014', 2021013966258, 'SARSG', '', '', '2021-03-23 08:06:00', '2021-03-23 08:26:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010015', 2021016153907, 'SARSG', '', '', '2021-03-31 16:40:00', '2021-03-31 17:00:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010016', 2021017448428, 'SARS2', '', '', '2021-04-05 08:00:00', '2021-04-05 18:36:00', 'NEGATIV', '', '2021-04-23 12:24:43', 0),
          ('2401010017', 2021017561852, 'SARSG', '', '', '2021-04-05 17:50:00', '2021-04-05 18:10:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010018', 2021018005363, 'SARSG', '', '', '2021-04-06 10:55:00', '2021-04-06 11:15:00', 'NEGATIV', 'SEI kommunal- og skoleindberetninger ', '2021-04-22 14:57:34', 0),
          ('2401010018', 2021018005364, 'SARSG', '', '', '2021-04-06 10:36:00', '2021-04-06 10:56:00', 'NEGATIV', 'SEI kommunal- og skoleindberetninger ', '2021-04-22 14:57:34', 0),
         */

        String messageId = UUID.randomUUID().toString();
        requestContext.setRequestContext(messageId);
        borgerContext.setContext(FMK_AUDIENCE, IDWS_CPR, IDWS_PID, null);
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(REQUEST_CPR);

        ((BindingProvider) servicePort).getRequestContext().put(MessageContext.HTTP_REQUEST_HEADERS, Collections.singletonMap("Xcachehit", Collections.singletonList("false")));

        GetCoronaPassResponse response = servicePort.getCoronaPass(request);
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        JAXB.marshal(response, os);
        log.info("GetCoronaPassResponse=" + os.toString());

        assertNotNull(response);
        Document responseDocument = documentInterceptor.getDocument();
        LibertyResponse libertyResponse = oioIdwsFactory.createResponse(responseDocument);
        assertNotNull(libertyResponse);

        // Notice: in order to validate response signature, we need to add the cert from "oio_idws_response_signing.jks" residing
        // in NSP docker image /pack/wildfly/modules/system/layers/base/dk/sds/nsp/accesshandler/main/ to our trustVault (trust.jks)
        libertyResponse.validateSignatureAndTrust(trustVault);
        assertEquals(messageId, libertyResponse.getRelatesTo());
    }

}
